#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>

#ifdef ENABLE_SYSLOG
#include <syslog.h>

#define	SYSLOG(x)	syslog( LOG_DEBUG, x )
#define	SYSLOG2(x,y)	syslog( LOG_DEBUG, x, y )

#else

#define	SYSLOG(x)	/* */
#define	SYSLOG2(x,y)	/* */

#endif

#define GETPW_NAMELEN 	255

char	GETPW_PWFILENAME[ GETPW_NAMELEN ] = "/etc/passwd";
FILE	*GETPW_PWFILEDESC;

/* The only real function I made.  :) */
void	getpw_setaltfile( const char *file )
{
	strcpy( GETPW_PWFILENAME, file );

	/* If the old file is open, then close it */
	if( GETPW_PWFILEDESC != NULL )
	{
		fclose( GETPW_PWFILEDESC );
		GETPW_PWFILEDESC = NULL;
	}
}

/* All the replacement funcs follow here */
struct passwd *getpwent(void)
{
#ifdef ENABLE_SYSLOG
	struct passwd	*pw;
#endif
	char	*envpw;
	char	*bin, *buf, *fpath;

	FILE	*conf;

	if( GETPW_PWFILEDESC == NULL )
	{
		/* If the ALTPASSWD env is set, use it always */
		envpw = getenv( "ALTPASSWD" );
		if( envpw != NULL )
		{
#ifdef ENABLE_PARANOID
			/* Validate that we are uid 0 */
			if( getuid() != 0 )
			{
				SYSLOG2( "Override attempted by non-root UID %d", getuid() );
				strcpy( GETPW_PWFILENAME, "/etc/passwd" );
			}
#endif
			strcpy( GETPW_PWFILENAME, envpw );
		}else
		{
			/* Otherwise, use /etc/passwd.switch */
			bin = (char *)malloc( 1024 );
			buf = (char *)malloc( 1024 );
			bzero( bin, 1024 );
			readlink( "/proc/self/exe", bin, 1023 );

			/* Now go through the config to resolve... */
			conf = fopen( "/etc/passwd.switch", "r" );
			if( conf )
			{
				while( conf )
				{
					if( fgets( buf, 1024, conf ) )
					{
						/* Clean any \n */
						if( buf[ strlen( buf ) - 1 ] == '\n' )
						{
							buf[ strlen( buf ) - 1 ] = 0;
						}

						fpath = strchr( buf, ':' );
						SYSLOG2( "Line read: %s", buf );

						if( fpath )
						{
							/* Null it here */
							*fpath = 0;
	
							/* Move to beg of string */
							fpath++;

							SYSLOG2( "buf: %s", buf );
							SYSLOG2( "fpath: %s", fpath );

							/* Check if stuff before : is in bin filename */
							if( ( strcmp( buf, "*" ) == 0 ) ||
								( strstr( bin, buf ) ) )
							{
								/* Match! */
								SYSLOG2( "Match for passwd file: %s", fpath );
								strcpy( GETPW_PWFILENAME, fpath );
								fclose( conf );
								conf = NULL;
							}
						}
					}else
					{
						SYSLOG( "No match found in /etc/passwd.switch for file" );
						fclose( conf );
						conf = NULL;
					}
				}
			}else
			{
				SYSLOG( "Unable to open: /etc/passwd.switch, I HAVE NO PASSWORD FILE!!" );
			}

			free( bin );
			free( buf );

			if( GETPW_PWFILENAME[ 0 ] == 0 )
			{
				printf( "Unable to locate a password file, bye now.\n" );
				exit( 0 );
			}
		}


		/* Need to open it first */
		GETPW_PWFILEDESC = fopen( GETPW_PWFILENAME, "r" );
		if( ! GETPW_PWFILEDESC )
		{
			printf( "Could not open file: '%s'\n", GETPW_PWFILENAME );
			exit( 1 );
		}

		SYSLOG2( "Opened pwfile '%s'", GETPW_PWFILENAME );
	}

#ifdef ENABLE_SYSLOG
	pw = fgetpwent( GETPW_PWFILEDESC );

	if( pw == NULL )
	{
		SYSLOG( "Got NULL ptr" );
	}else
	{
		SYSLOG2( "Got username: '%s'", pw -> pw_name );
	}

	return( pw );
#else
	return( fgetpwent( GETPW_PWFILEDESC ) );
#endif
}

void	setpwent( void )
{
	/* Rewind the ptr */
	if( GETPW_PWFILEDESC != NULL )
	{
		/* Just for now, this will reset it since it will re-open it */
		fclose( GETPW_PWFILEDESC );
		GETPW_PWFILEDESC = NULL;
	}
}

void	endpwent( void )
{
	setpwent();
}

struct passwd *getpwnam(const char * name)
{
	struct passwd	*pw;

	/* Rewind */
	setpwent();

	/* Now scan on through... */
	while( 1 )
	{
		pw = getpwent();

		if( ( pw == NULL ) ||
			( strcmp( name, pw -> pw_name ) == 0 ) )
		{
			return( pw );
		}
	}
}

struct passwd *getpwuid(uid_t uid)
{
	struct passwd	*pw;

	/* Rewind */
	setpwent();

	/* Now scan on through... */
	while( 1 )
	{
		pw = getpwent();

		if( ( pw == NULL ) ||
			( uid == pw -> pw_uid ) )
		{
			return( pw );
		}
	}
}

int getpw(uid_t uid, char *buf)
{
	/* Make the pwent */
	struct passwd	*pw;

	pw = getpwuid( uid );
	if( pw == NULL )
	{
		return( -1 );
	}

	sprintf( buf, "%s:%s:%d:%d:%s:%s:%s",
		pw -> pw_name,
		pw -> pw_passwd,
		pw -> pw_uid,
		pw -> pw_gid,
		pw -> pw_gecos,
		pw -> pw_dir,
		pw -> pw_shell );

	return( 0 );
}

int putpwent(const struct passwd *p, FILE *stream)
{
	/* Not supported */
	return( -1 );
}
